/* -*- Mode: C; c-basic-offset: 8; tab-width: 8 -*- */
/*
 *  AEWPHelper.c
 *
 *  Created by Erik Larsson on 2009-09-04.
 *  Copyright 2009 Tuxera Ltd. All rights reserved.
 *
 *  This helper tool executes another tool and filters its output to return it
 *  in a structured form suitable for interpretation by code executing
 *  AuthorizationExecuteWithPrivileges.
 *
 *  AuthorizationExecuteWithPrivileges is a nasty function which doesn't give
 *  much control to the caller. For instance there is no way to peek at the
 *  process ID or exit status of the executed process, and you cannot get access
 *  to its error output (stderr), so this tool "multiplexes" all this
 *  information into the standard output stream using a simple line-based
 *  protocol.
 *
 *  The following messages are part of the protocol:
 *
 *  PID: <int pid>\n
 *    The process ID of the currently running AEWPHelper process.
 *
 *  CHILDPID: <int pid>\n
 *    The process ID of the child process, where the actual tool is executed.
 *
 *  STDOUT: <unsigned int length>\n
 *  <data>\n
 *    A chunk of stdout data. The data does not, as in the other cases, proceed
 *    until a line ending. Instead the length of the data is given, and the
 *    client should only read the amount of data specified by length. After that
 *    length of data, a line ending is still required to signal the end of that
 *    command.
 *
 *  STDERR: <unsigned int length>\n
 *  <data>\n
 *    A chunk of stderr data. See STDOUT.
 *
 *  EXITVALUE: <int value>\n
 *    The exit value of the child process. This indicates that the child is
 *    finished with its operation.
 *
 *  ERROR: <int errorcode> <string message>\n
 *    Indicates that an internal error was encountered in AEWPHelper, such as
 *    "fork() failed" or "couldn't execve the external tool".
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
//#include <syslog.h>

#define EXIT_VALUE_INVALID_ARGUMENT_COUNT       1
#define EXIT_VALUE_SETUID_FAILED                2
#define EXIT_VALUE_FORK_FAILED                  3
#define EXIT_VALUE_OPEN_SINK_PIPE_FAILED        4
#define EXIT_VALUE_REALLOC_MEMBUF_FAILED        5
#define EXIT_VALUE_UNCLEAN_PROCESS_EXIT         6
#define EXIT_VALUE_WAITPID_FAILED               7
#define EXIT_VALUE_CREATE_OUTPIPE_FAILED        8
#define EXIT_VALUE_CREATE_ERRPIPE_FAILED        9
#define EXIT_VALUE_READ_STDOUT_FAILED           10
#define EXIT_VALUE_READ_STDERR_FAILED           11
//#define EXIT_VALUE_

static pthread_mutex_t emit_mutex;

/** Emits an error message and then exits the program with the specified error code. (From 'errors' there is no recovery...) */
#define emitError(errorCode, fmt, ...) do { fprintf(stdout, "ERROR: %d " fmt "\n", errorCode, ##__VA_ARGS__); exit(errorCode); } while(0)

#define emitData(streamname, length, data) do { if(pthread_mutex_lock(&emit_mutex) != 0) { emitError(-1, "Failed to lock emit_mutex!"); } fprintf(stdout, "%s: %zd\n", streamname, length); fwrite(data, length, 1, stdout); fprintf(stdout, "\n"); fflush(stdout); if(pthread_mutex_unlock(&emit_mutex) != 0) { emitError(-1, "Failed to unlock emit_mutex!"); } } while(0)

#define emitPid(pid) do { fprintf(stdout, "PID: %d\n", pid); } while(0)
#define emitChildPid(pid) do { fprintf(stdout, "CHILDPID: %d\n", pid); } while(0)
#define emitStdout(length, data) emitData("STDOUT", length, data)
#define emitStderr(length, data) emitData("STDERR", length, data)
#define emitExitValue(exitValue) do { fprintf(stdout, "EXITVALUE: %d\n", exitValue); } while(0)

#define CFLog(args...) do { CFStringRef string = CFStringCreateWithFormat(kCFAllocatorDefault, NULL, args); CFShow(string); CFRelease(string); } while(0)

#if DEBUG
//#define LOG_INIT() openlog("AEWPHelper", LOG_CONS | LOG_PID, LOG_USER)
////#define LOG_INIT() openlog("AEWPHelper", LOG_CONS | LOG_PID | LOG_PERROR, LOG_USER)
//#define LOG_DEBUG(fmt, ...) syslog(LOG_NOTICE, fmt, __VA_ARGS__)
//#define LOG_CLEANUP() closelog()
#define LOG_INIT()
#define LOG_DEBUG(fmt, ...) CFLog(fmt, ##__VA_ARGS__)
#define LOG_CLEANUP()
#else
#define LOG_INIT()
#define LOG_DEBUG(fmt, ...)
#define LOG_CLEANUP()
#endif

#define MEMBUF_SIZE 1 // 64*1024

static void* process_stderr(void *arg) {
	int pipe = *((int*)arg);
	int errEOF = 0;
	size_t membufSize = MEMBUF_SIZE;
	char *memBuf = malloc(membufSize);

	if(memBuf == NULL)
		emitError(-1, "Couldn't malloc membuf: %d (%s)",
			  errno, strerror(errno));

	while(!errEOF) {
		ssize_t bytesRead;

		bytesRead = read(pipe, memBuf, membufSize);
		if(bytesRead > 0) {
			emitStderr(bytesRead, memBuf);
		}
		else if(bytesRead == 0) {
			errEOF = 1;
		}
		else {
			emitError(EXIT_VALUE_READ_STDERR_FAILED,
				  "Failed to read from sink "
				  "side of stderr pipe: %d "
				  "(%s)",
				  errno, strerror(errno));
		}
	}

	free(memBuf);

	return NULL;
}

static void* process_stdout(void *arg) {
	int pipe = *((int*)arg);
	int outEOF = 0;
	size_t membufSize = MEMBUF_SIZE;
	char *memBuf = malloc(membufSize);

	if(memBuf == NULL)
		emitError(-1, "Couldn't malloc membuf: %d (%s)",
			  errno, strerror(errno));

	while(!outEOF) {
		ssize_t bytesRead;

		bytesRead = read(pipe, memBuf,
				 membufSize);
		if(bytesRead > 0) {
			emitStdout(bytesRead, memBuf);
		}
		else if(bytesRead == 0) {
			outEOF = 1;
		}
		else {
			emitError(EXIT_VALUE_READ_STDOUT_FAILED,
				  "Failed to read from sink "
				  "side of stdout pipe: %d "
				  "(%s)",
				  errno, strerror(errno));
		}
	}

	free(memBuf);

	return NULL;
}

void exec_utility(int argc, char **argv) {
	char *path = argv[0];
	int outpipe[2];
	int errpipe[2];

	if(pipe(outpipe) != 0) {
		emitError(EXIT_VALUE_CREATE_OUTPIPE_FAILED,
			  "Couldn't create pipe for capturing stdout. errno: "
			  "%d (%s)", errno, strerror(errno));
	}
	if(pipe(errpipe) != 0) {
		emitError(EXIT_VALUE_CREATE_ERRPIPE_FAILED,
			  "Couldn't create pipe for capturing stderr. errno: "
			  "%d (%s)", errno, strerror(errno));
	}

	LOG_DEBUG("forking...");
	pid_t childPid = fork();
	if(childPid == -1) {
		LOG_DEBUG("ERROR: fork failed!");
		emitError(EXIT_VALUE_FORK_FAILED, "Could not fork()!");
	}
	else if(childPid == 0) {
		LOG_DEBUG("fork succeeded - now in child uid=%d euid=%d",
			  getuid(), geteuid());
		//sleep(3);

		// Close the sink ends of the pipes.
		close(outpipe[0]);
		close(errpipe[0]);

		fclose(stdout);
		dup2(outpipe[1], STDOUT_FILENO);
		dup2(errpipe[1], STDERR_FILENO);

		char **childArgs = argv;
		LOG_DEBUG("invoking execv...");
		execv(path, childArgs);
		LOG_DEBUG("ERROR: execv failed with errno=%d", errno);
		fprintf(stderr, "[AEWPHelper] execv failed with errno: %d "
			"(%s)\n", errno, strerror(errno));
		LOG_CLEANUP();
		exit(-1);
	}
	else {
		int status;

		emitChildPid(childPid);
		LOG_DEBUG("fork succeeded - now in parent");

		// Close the source ends of the pipes.
		close(outpipe[1]);
		close(errpipe[1]);

		LOG_DEBUG("reading output from child process...");

		/* Create the two threads responsible for reading the process
		 * stdin and stderr streams. */
		pthread_t outThread, errThread;
		pthread_create(&outThread, NULL, process_stdout, &outpipe[0]);
		pthread_create(&errThread, NULL, process_stderr, &errpipe[0]);

		pthread_join(outThread, NULL);
		pthread_join(errThread, NULL);

		LOG_DEBUG("finished reading output from child process...");

		// Close the sink ends of the pipes.
		close(outpipe[0]);
		close(errpipe[0]);

		LOG_DEBUG("waiting for child processes...");
		pid_t waitres = waitpid(childPid, &status, 0);
		LOG_DEBUG("waiting complete! waitres=%d", waitres);
		if(waitres > 0) {
			if(WIFEXITED(status)) {
				int exitValue = WEXITSTATUS(status);
				emitExitValue(exitValue);
			}
			else
				emitError(EXIT_VALUE_UNCLEAN_PROCESS_EXIT,
					  "Process did not exit cleanly "
					  "(crash?) status=0x%X", status);
		}
		else
			emitError(EXIT_VALUE_WAITPID_FAILED, "waitpid failed! "
				  "waitres=%d", waitres);
	}
}

int main(int argc, char **argv) {
	//int retval = 1;

	if(pthread_mutex_init(&emit_mutex, NULL) != 0)
		emitError(-1, "pthread_mutex_init failed!");

	emitPid(getpid());

	LOG_INIT();
	LOG_DEBUG("Started. uid=%d euid=%d", getuid(), geteuid());

	if(setuid(0) != 0)
		emitError(EXIT_VALUE_SETUID_FAILED, "Could not setuid(0)");
	else {
		LOG_DEBUG("After setuid: uid=%d euid=%d", getuid(), geteuid());
		if(argc < 2) {
			emitError(EXIT_VALUE_INVALID_ARGUMENT_COUNT,
				  "Invalid number of arguments. (got: %d, "
				  "minimum required: 1)", (argc-1));
		}
		else {
			exec_utility(argc-1, argv+1);
		}
	}

	LOG_CLEANUP();

	pthread_mutex_destroy(&emit_mutex);

	return 0;
}
