/*-
 * Copyright (C) 2007-2009 Erik Larsson
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

//
//  DriverUtil.m
//  NTFS-3G_PrefPane
//
//  Created by Erik Larsson on 2008-11-23.
//

#import <sys/utsname.h>
//#import <syslog.h>

#import "DriverUtil.h"
#import "CommonDriverStrings.h"
#import "AEWPHelperClient.h"

@implementation NTFS3GDriverUtil

#if 0
#define DebugLog(fmt, args...) NSLog(fmt, ##args)
#else
#define DebugLog(fmt, args...)
#endif

typedef struct {
	NSMutableString *stdoutString;
	NSMutableArray *stdoutArray;
	NSMutableString *stderrString;
	NSMutableArray *stderrArray;
} AddResultContext;

static void addResultCallback(char *stdoutChunk, size_t stdoutChunkLen, NSMutableString **curStringPtr, NSMutableArray *lineArray)
{
	NSString *tempString = nil;
	char *curPtr = stdoutChunk;

	if(!stdoutChunk) {
		NSLog(@"Error: stdoutChunk == NULL");
		return;
	}
	else if(!curStringPtr) {
		NSLog(@"Error: curStringPtr == NULL");
		return;
	}
	else if(!lineArray) {
		NSLog(@"Error: lineArray == NULL");
		return;
	}

	while((curPtr - stdoutChunk) < stdoutChunkLen) {
		char *endPtr = curPtr;
		while(*endPtr != '\n' && (endPtr - stdoutChunk) < stdoutChunkLen)
			++endPtr; // Fast-forward to next newline.

		tempString = [[NSString alloc] initWithBytesNoCopy: curPtr
													length: (endPtr-curPtr)
												  encoding: NSUTF8StringEncoding
											  freeWhenDone: NO];

		if(!tempString) {
			NSLog(@"Failed to decode string \"%.*s\" as UTF-8.", (curPtr - endPtr), curPtr);
		}
		else {
			if(*curStringPtr) {
				NSString *newTempString = [[NSString alloc] initWithFormat: @"%@%@", *curStringPtr, tempString];
				[*curStringPtr release];
				[tempString release];
				tempString = newTempString;
				*curStringPtr = nil;
			}

			if((endPtr - stdoutChunk) < stdoutChunkLen)
				[lineArray addObject: tempString];
			else {
				NSMutableString *outgoing = [[NSMutableString alloc] initWithString: tempString];
				*curStringPtr = outgoing;
			}

			[tempString release];
		}

		curPtr = endPtr;
		if((curPtr - stdoutChunk) < stdoutChunkLen)
			++curPtr; // Skip newline character previously encountered.
	}
}

static void helperToolStdoutCallback(void *data, char *stdoutChunk, size_t stdoutChunkLen)
{
	AddResultContext* context = (AddResultContext*) data;
	addResultCallback(stdoutChunk, stdoutChunkLen, &context->stdoutString, context->stdoutArray);
}

static void helperToolStderrCallback(void *data, char *stderrChunk, size_t stderrChunkLen)
{
	AddResultContext* context = (AddResultContext*) data;
	addResultCallback(stderrChunk, stderrChunkLen, &context->stderrString, context->stderrArray);
}

/**
 * Invokes a helper tool with elevated privileges and parses the standard responses that our "void" helper tools
 * adhere to.
 */
+ (NSString*)invokeHelperTool: (NSString*)toolName withResult: (NSArray**)resultPointer withBundle: (NSBundle*)bundle withAuth: (AuthorizationRef)auth, ...
{
	NSString *errorMessage = nil;
	//openlog("NTFS3GOperations", LOG_CONS | LOG_PID, LOG_USER);

	NSString *aewpHelperPath = [bundle pathForResource: TOOL_AEWPHELPER ofType: nil];
	NSString *toolPath = [bundle pathForResource: toolName ofType: nil];
	if(toolPath == nil) {
		errorMessage = [[[NSString alloc] initWithFormat: @"[invokeHelperTool] Could not get bundle resource for %@.\nbundle bundlePath: %@",
						 toolName, [bundle bundlePath]] autorelease];
	}
	else if(aewpHelperPath == nil) {
		errorMessage = [[[NSString alloc] initWithFormat: @"[invokeHelperTool] Could not get bundle resource for %@.\nbundle bundlePath: %@",
						 TOOL_AEWPHELPER, [bundle bundlePath]] autorelease];
	}
	else {
		int myArgumentsLength = 2;
		char **myArguments = malloc(myArgumentsLength*sizeof(char*));
		if(myArguments == NULL)
			errorMessage = @"[invokeHelperTool] Malloc error while allocating initial myArguments!";
		else {
			int cToolPathSize = [toolPath length]*3 + 1;
			char *cToolPath = malloc(cToolPathSize);
			if(!cToolPath)
				errorMessage = @"[invokeHelperTool] Malloc of cToolPath failed.";
			else {
				[toolPath getCString: cToolPath maxLength: cToolPathSize encoding: NSUTF8StringEncoding];
				myArguments[0] = cToolPath;
				myArguments[1] = NULL;
			}
		}
		
		if(errorMessage == nil) {
			// Unfold the varargs into myArguments
			id curObject;
			va_list argumentList;
			va_start(argumentList, auth);          // Start scanning for arguments after auth.
			while(curObject = va_arg(argumentList, NSString*)) { // As many times as we can get an argument of type "NSString*"
				if([curObject isKindOfClass: [NSString class]]) {
					// Make room for yet another argument in myArguments.
					char **newArguments = realloc(myArguments, (++myArgumentsLength)*sizeof(char*));
					if(newArguments == NULL) {
						errorMessage = @"[invokeHelperTool] Could not realloc myArguments while trying to parse varargs!";
						break;
					}
					else {
						NSString *str = (NSString*)curObject;
						
						// realloc went fine, so we can assign the result to myArguments
						myArguments = newArguments;
						
						// Allocate sufficient space to hold the UTF-8 form of the argument.
						int cstrArgMaxLength = [str length]*3; // length is UTF-16 unit count. 1 UTF-16 unit -> 1-3 UTF-8 units
						char *cstrArg = malloc(cstrArgMaxLength + 1);
						if(cstrArg == NULL) {
							errorMessage = @"[invokeHelperTool] Could not malloc cstrArg while trying to parse varargs!";
							break;
						}
						
						// Convert the current argument to UTF-8 form.
						[str getCString: cstrArg maxLength: cstrArgMaxLength + 1 encoding: NSUTF8StringEncoding];
						
						// Store in myArguments.
						myArguments[myArgumentsLength-2] = cstrArg;
						myArguments[myArgumentsLength-1] = NULL;
					}
				}
				else {
					errorMessage = [[[NSString alloc] initWithFormat: @"[invokeHelperTool] Encountered unexpected type in varargs list!\n"
									 "Expected: NSString, Got: %@ (myArguments length: %d)", curObject, myArgumentsLength] autorelease];
					//errorMessage = @"[invokeHelperTool] Encountered unexpected type in varargs list!";
					break;
				}
			}
			va_end(argumentList);
		}

		if(errorMessage == nil) {
			if(0) {
				int i;
				DebugLog(@"Invoking AuthorizationExecuteWithPrivileges:");
				DebugLog(@"  aewpHelperPath: \"%@\"", aewpHelperPath);
				DebugLog(@"  myArguments:");
				for(i = 0; i < myArgumentsLength; ++i)
					DebugLog(@"    \"%s\"", myArguments[i]);
			}

			FILE *aewpHelperPipe = NULL;
			OSStatus status = AuthorizationExecuteWithPrivileges(auth, [aewpHelperPath UTF8String],
																 kAuthorizationFlagDefaults, myArguments, &aewpHelperPipe);
			if(status != errAuthorizationSuccess)
				errorMessage = @"AuthorizationExecuteWithPrivileges failed.";
			else {
				AEWPHelperResults helperResults;
				AddResultContext context;
				context.stdoutString = nil;
				context.stderrString = nil;
				context.stdoutArray = [[NSMutableArray alloc] initWithCapacity: 1];
				context.stderrArray = [[NSMutableArray alloc] initWithCapacity: 1];

				if(readAEWPHelperOutput(aewpHelperPipe, &helperResults, &context,
										helperToolStdoutCallback, helperToolStderrCallback)) {
					if(helperResults.hasPid) {
						int stat_loc;
						waitpid(helperResults.pid, &stat_loc, 0);
					}

					if(context.stdoutString) {
						DebugLog(@"stdout output did not end with a newline. Trailing string: \"%@\"", context.stdoutString);
						[context.stdoutArray addObject: context.stdoutString];
						[context.stdoutString release];
						context.stdoutString = nil;
					}
					if(context.stderrString) {
						DebugLog(@"stderr output did not end with a newline. Trailing string: \"%@\"", context.stderrString);
						[context.stderrArray addObject: context.stderrString];
						[context.stderrString release];
						context.stderrString = nil;
					}

					DebugLog(@"  Finished reading AEWPHelper output.");
					if(helperResults.hasPid)
						DebugLog(@"    helperResults.pid = %d", helperResults.pid);
					if(helperResults.hasChildpid)
						DebugLog(@"    helperResults.childpid = %d", helperResults.childpid);
					if(helperResults.hasErrorCode)
						DebugLog(@"    helperResults.errorCode = %d", helperResults.errorCode);
					if(helperResults.hasExitValue)
						DebugLog(@"    helperResults.exitValue = %d", helperResults.exitValue);
					if(0) {
						int i;
						DebugLog(@"    stdout (lines=%d):", [context.stdoutArray count]);
						for(i = 0; i < [context.stdoutArray count]; ++i)
							DebugLog(@"      %d: %@", i + 1, [context.stdoutArray objectAtIndex: i]);
						DebugLog(@"    stderr (lines=%d):", [context.stderrArray count]);
						for(i = 0; i < [context.stderrArray count]; ++i)
							DebugLog(@"      %d: %@", i + 1, [context.stderrArray objectAtIndex: i]);
					}

					if(helperResults.hasExitValue && helperResults.exitValue == 0) {
						DebugLog(@"Successful execution.");
						if(resultPointer)
							*resultPointer = context.stdoutArray;
						[context.stderrArray release];
						context.stdoutArray = nil;
						context.stderrArray = nil;
					}
					else {
						int i;
						NSMutableString *errorMessageBuilder = [[NSMutableString alloc] initWithCapacity: 0];

						DebugLog(@"Unsuccessful execution:");
						DebugLog(@"  hasExitValue = %d", helperResults.hasExitValue);
						if(helperResults.hasExitValue)
							DebugLog(@"  exitValue = %d", helperResults.exitValue);

						for(i = 0; i < [context.stderrArray count]; ++i)
							[errorMessageBuilder appendFormat: @"%@\n", [context.stderrArray objectAtIndex: i]];
						DebugLog(@"Returning error message: %p (\"%@\")", errorMessageBuilder, errorMessageBuilder);
						[errorMessageBuilder autorelease];
						errorMessage = errorMessageBuilder;
						[context.stdoutArray release];
						[context.stderrArray release];
						context.stdoutArray = nil;
						context.stderrArray = nil;
					}
				}

				fclose(aewpHelperPipe);
			}
		}

		if(myArguments != NULL) {
			// Free all memory allocated for myArguments
			int ptr = 0;
			char *curArg;
			while((curArg = myArguments[ptr++]) != NULL) {
				free(curArg);
			}
			free(myArguments);
		}
	}

	//closelog();
	return errorMessage;
}

/*+ (int)setDebugLoggingState:(BOOL)enabled withAuth:(AuthorizationRef)auth
{
	FILE *myCommunicationsPipe = NULL;

	char *myArguments[] = {(enabled?"2":"1"), NULL};
	
	//char myPath[MAXPATHLEN];
	OSStatus status =
		AuthorizationExecuteWithPrivileges(auth, "/System/Library/Filesystems/ntfs-3g.fs/Support/setloglevel.sh",
				       kAuthorizationFlagDefaults, myArguments, &myCommunicationsPipe);
	
	fclose(myCommunicationsPipe);
  
	return status;
}*/


+ (NSString*)listMountpoints:(AuthorizationRef)auth toMountPointArray:(NSMutableArray*)mountpointArray toStatusArray: (NSMutableArray*)statusArray withBundle:(NSBundle*)bundle
{
	struct utsname systemUname;
	uname(&systemUname);
	
	NSString *helperTool;	
	if(memcmp(systemUname.release, "8.", 2) == 0) {
		helperTool = TOOL_LISTFSMOUNTSTIGER;
	}
	else {
		helperTool = TOOL_LISTFSMOUNTSLEOPARD;
	}
	
	NSArray *result = nil;
	NSString *errorMessage = [DriverUtil invokeHelperTool: helperTool
											   withResult: &result
											   withBundle: bundle
												 withAuth: auth,
							  NULL];
	
	if(errorMessage == nil) {
		int i;
		for(i = 0; i+1 < [result count]; i += 2) {
			[mountpointArray addObject: [result objectAtIndex: i]];
			[statusArray addObject: [result objectAtIndex: i+1]];
		}
	}
	if(result != nil)
		[result release];
	
	return errorMessage;
}

+ (NSString*)getPreferenceValue: (NSString*)key outValue: (int*)output
{
	NSString *errorMessage = nil;
	NSDictionary *driverSettingsDict = [NSDictionary dictionaryWithContentsOfFile: DRIVER_SETTINGS_PLIST_PATH];
	if(driverSettingsDict == nil) {
		errorMessage = @"Could not read settings dictionary.";
	}
	else {
		id value = [driverSettingsDict valueForKey: key];
		if(value == nil)
			errorMessage = [[[NSString alloc] initWithFormat: @"No value for key: %@", key] autorelease];
		else if([value isKindOfClass: [NSNumber class]]) {
			NSNumber *nsn = (NSNumber*)value;
			*output = [nsn intValue];
		}
		else if([value isKindOfClass: [NSString class]]) {
			NSString *nsn = (NSString*)value;
			*output = [nsn intValue];
		}
		else
			errorMessage = [[[NSString alloc] initWithFormat: @"Invalid value class: %@", [value class]] autorelease];
	}
	
	return errorMessage;
}

+ (NSString*)getPreferenceValue: (NSString*)key outValueAsLongLong: (long long*)output
{
	NSString *errorMessage = nil;
	NSDictionary *driverSettingsDict = [NSDictionary dictionaryWithContentsOfFile: DRIVER_SETTINGS_PLIST_PATH];
	if(driverSettingsDict == nil) {
		errorMessage = @"Could not read settings dictionary.";
	}
	else {
		id value = [driverSettingsDict valueForKey: key];
		if(value == nil)
			errorMessage = [[[NSString alloc] initWithFormat: @"No value for key: %@", key] autorelease];
		else if([value isKindOfClass: [NSNumber class]]) {
			NSNumber *nsn = (NSNumber*)value;
			*output = [nsn longLongValue];
		}
		else if([value isKindOfClass: [NSString class]]) {
			NSString *nss = (NSString*)value;
			if(sscanf([nss UTF8String], "%llu", output) != 1)
				errorMessage = @"scanf failed";
		}
		else
			errorMessage = [[[NSString alloc] initWithFormat: @"Invalid value class: %@", [value class]] autorelease];
	}
	
	return errorMessage;
}

+ (NSString*)setPreferenceKey: (NSString*)key toValue: (NSString*)value withBundle: (NSBundle*)bundle withAuth: (AuthorizationRef)auth
{
	return [DriverUtil invokeHelperTool: TOOL_SETPLISTVALUE
							 withResult: NULL 
							 withBundle: bundle
							   withAuth: auth,
			DRIVER_SETTINGS_PLIST_PATH,
			key,
			value,
			NULL];
}

+ (NSString*)createEmptyFile: (NSString*)filename withBundle: (NSBundle*)bundle withAuth: (AuthorizationRef)auth
{
	return [DriverUtil invokeHelperTool: TOOL_CREATEEMPTYFILE withResult: NULL withBundle: bundle withAuth: auth, filename, NULL];
}


+ (NSString*)deleteFile: (NSString*)filename withBundle: (NSBundle*)bundle withAuth: (AuthorizationRef)auth
{
	return [DriverUtil invokeHelperTool: TOOL_REMOVEFILE withResult: NULL withBundle: bundle withAuth: auth, filename, NULL];
}

+ (NSString*)existsFile: (NSString*)filename withResult: (BOOL*)result withBundle: (NSBundle*)bundle withAuth: (AuthorizationRef)auth
{
	NSArray *resultArray = nil;
	NSString *errorMessage = [DriverUtil invokeHelperTool: TOOL_EXISTSFILE withResult: &resultArray withBundle: bundle withAuth: auth, filename, NULL];

	if(resultArray != nil && [resultArray count] == 1 && [[resultArray objectAtIndex: 0] isEqual: @"TRUE"]) {
		*result = TRUE;
	}
	else {
		*result = FALSE;
	}
	/* //DEBUG
	if(errorMessage == nil)
		errorMessage = [NSString stringWithFormat: @"resultArray (length=%d):\n%@", [resultArray count], [resultArray objectAtIndex: 0]];
	*/
	
	return errorMessage;
}

+ (NSString*)setControlFile: (NSString*)filename inDirectory: (NSString*)dirname toState: (BOOL)existent withBundle: (NSBundle*)bundle withAuth: (AuthorizationRef)auth
{
	NSString *errorMessage;
	NSString *controlFilePath = [[NSString alloc] initWithFormat: @"%@/%@", dirname, filename];
	
	if(existent == TRUE) {
		errorMessage = [DriverUtil invokeHelperTool: TOOL_ENSUREDIRECTORY
										 withResult: NULL
										 withBundle: bundle
										   withAuth: auth, dirname, NULL];
		if(errorMessage == nil) {
			errorMessage = [DriverUtil invokeHelperTool: TOOL_CREATEEMPTYFILE
											 withResult: NULL
											 withBundle: bundle
											   withAuth: auth, controlFilePath, NULL];
		}
	}
	else if(existent == FALSE) {
		errorMessage = [DriverUtil invokeHelperTool: TOOL_REMOVEFILE
										 withResult: NULL
										 withBundle: bundle
										   withAuth: auth, controlFilePath, NULL];
	}
	else
		errorMessage = @"Internal error: Parameter 'existent' is neither TRUE nor FALSE!";
	
	[controlFilePath release];
	return errorMessage;	
}

+ (NSString*)getMountFlags: (NSString*)filename toArray: (NSMutableArray*)optionArray withBundle: (NSBundle*)bundle withAuth: (AuthorizationRef)auth
{
	NSArray *resultArray = nil;
	NSString *errorMessage = [DriverUtil invokeHelperTool: TOOL_GETMOUNTFLAGS
											   withResult: &resultArray
											   withBundle: bundle
												 withAuth: auth, filename, NULL];
	
	if(errorMessage == nil)
		[optionArray addObjectsFromArray: resultArray];
	if(resultArray != nil)
		[resultArray release];
	return errorMessage;
}

+ (NSString*)invokeUninstallScriptWithBundle: (NSBundle*)bundle withAuth: (AuthorizationRef)auth toOutputArray:(NSMutableArray*)outputArray
{
	NSMutableArray *result = nil;
	NSString *errorMessage = [DriverUtil invokeHelperTool: TOOL_PROCESSRUNNER
											   withResult: &result
											   withBundle: bundle
												 withAuth: auth, /*@"/bin/bash",*/ DRIVER_UNINSTALL_SCRIPT_PATH, NULL];
	if(errorMessage == nil) {
		int i;
		for(i = 0; i < [result count]; ++i) {
			[outputArray addObject: [result objectAtIndex: i]];
		}
	}
	if(result != nil)
		[result release];
	
	return errorMessage;
}

@end
