/*-
 * Copyright (C) 2007-2009 Erik Larsson
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#import "NTFS3GPrefPaneController.h"
#import "DriverUtil.h"
#import "CommonDriverStrings.h"

#define LOG_DEBUG(...)

@implementation NTFS3GPrefPaneController

- (void)alertDidEnd:(NSAlert *)alert returnCode:(int)returnCode
		contextInfo:(void *)contextInfo
{
}

- (void)messageWithTitle: (NSString*)titleText
			withInfoText: (NSString*)infoText
		  withAlertStyle: (NSAlertStyle)alertStyle
			   forWindow: (NSWindow*)messageParentWindow
		   modalDelegate: (id)modalDelegate
		  didEndSelector: (SEL)alertDidEndSelector
{
	NSAlert *alert = [[[NSAlert alloc] init] autorelease];
	
	[alert addButtonWithTitle: @"OK"];
	
	[alert setMessageText: titleText];
	
	[alert setInformativeText: infoText];
	
	[alert setAlertStyle: alertStyle];
	
	[alert beginSheetModalForWindow: messageParentWindow
					  modalDelegate: modalDelegate
					 didEndSelector: alertDidEndSelector
						contextInfo: nil];
}

- (void)messageWithTitle: (NSString*)titleText
			withInfoText: (NSString*)infoText
		  withAlertStyle: (NSAlertStyle)alertStyle
			   forWindow: (NSWindow*)messageParentWindow
{
	[self messageWithTitle:titleText
			  withInfoText:infoText
			withAlertStyle:alertStyle
				 forWindow:messageParentWindow
			 modalDelegate:self
			didEndSelector:@selector(alertDidEnd:returnCode:contextInfo:)];
}

- (void)errorMessage:(NSString*)message forWindow:(NSWindow*)messageParentWindow
{
	[self messageWithTitle: @"Error" withInfoText: message withAlertStyle: NSWarningAlertStyle forWindow: messageParentWindow];
}

- (void)infoMessageWithTitle: (NSString*)titleText withInfoText: (NSString*)infoText forWindow: (NSWindow*)messageParentWindow
{
	[self messageWithTitle: titleText withInfoText: infoText withAlertStyle: NSInformationalAlertStyle forWindow: messageParentWindow];
}

// From: http://vgable.com/blog/2008/10/05/restarting-your-cocoa-application/
- (void) restartOurselves
{
	LOG_DEBUG(@"restartOurselves invoked");
	NSString *killArg1AndOpenArg2Script = @"kill -TERM $1\nopen \"$2\"";
	
	//NSTask needs its arguments to be strings
	NSString *ourPID = [NSString stringWithFormat:@"%d",
						[[NSProcessInfo processInfo] processIdentifier]];
	
	LOG_DEBUG(@"  ourPID=\"%@\"", ourPID);
	//this will be the path to the .app bundle,
	//not the executable inside it; exactly what `open` wants
	NSString * pathToUs = [[NSBundle mainBundle] bundlePath];
	LOG_DEBUG(@"  pathToUs=\"%@\"", pathToUs);
	
	NSArray *shArgs = [NSArray arrayWithObjects:@"-c", // -c tells sh to execute the next argument, passing it the remaining arguments.
					   killArg1AndOpenArg2Script,
					   @"", //$0 path to script (ignored)
					   ourPID, //$1 in restartScript
					   pathToUs, //$2 in the restartScript
					   nil];
	
	NSTask *restartTask = [[NSTask alloc] init];
	[restartTask setLaunchPath: @"/bin/sh"];
	[restartTask setArguments: shArgs];
	
	NSPipe *errPipe = [NSPipe pipe];
	NSPipe *outPipe = [NSPipe pipe];
	
	[restartTask setStandardError: errPipe];
	[restartTask setStandardOutput: outPipe];
	
	[restartTask launch];
	
	NSData *errData = [[errPipe fileHandleForReading] readDataToEndOfFile];
	NSData *outData = [[outPipe fileHandleForReading] readDataToEndOfFile];	
		
	[restartTask waitUntilExit]; //wait for killArg1AndOpenArg2Script to finish
	
	NSLog(@"*** ERROR: %@ should have been terminated, but we are still running! return value from task: %d", pathToUs, [restartTask terminationStatus]);
	NSLog(@"task stderr: \"%@\"", [[[NSString alloc] autorelease] initWithData: errData encoding: NSUTF8StringEncoding]);
	NSLog(@"task stdout: \"%@\"", [[[NSString alloc] autorelease] initWithData: outData encoding: NSUTF8StringEncoding]);
	
	[restartTask release];
}

- (void)refreshGlobalOptions
{
	NSString *errorMessage = nil;
	NSMutableString *errorMessages = [[NSMutableString alloc] initWithCapacity: 0];
	int disableDriverValue = -1;
	int logLevelValue = -1;
	int mountRdisksValue = -1;
	int useUblioValue = -1;
	int useNfconvValue = -1;
	int alwaysRecoverValue = -1;
	
	BOOL existsUninstallScript = FALSE;
	
	LOG_DEBUG("Entering refreshGlobalOptions...");
	
	errorMessage = [DriverUtil getPreferenceValue: PLIST_KEY_DISABLE_DRIVER outValue: &disableDriverValue];
	LOG_DEBUG("Result for %@: %d\n", PLIST_KEY_DISABLE_DRIVER, disableDriverValue);
	LOG_DEBUG("  errorMessage=0x%p", errorMessage);
	if(errorMessage != nil) {
		disableDriverValue = 0;
		//[driverEnabledStatusField setStringValue: @"Unknown"];
		//[toggleDriverButton setEnabled: FALSE];
		//[errorMessages appendFormat: @"%@: %@\n", PLIST_KEY_DISABLE_DRIVER, errorMessage];
	}
	if(disableDriverValue == 0) {
		[driverEnabledStatusField setStringValue: @"Enabled"];
		[toggleDriverButton setTitle: @"Disable"];
		[self setGlobalOptionComponentsEnabled: TRUE];
	}
	else if(disableDriverValue == 1) {
		[driverEnabledStatusField setStringValue: @"Disabled"];
		[toggleDriverButton setTitle: @"Enable"];
		[self setGlobalOptionComponentsEnabled: FALSE];
	}
	else
		[errorMessages appendFormat: @"%@ value (%d) invalid.\n", PLIST_KEY_DISABLE_DRIVER, disableDriverValue];
	
	errorMessage = [DriverUtil getPreferenceValue: PLIST_KEY_LOGGING_LEVEL outValue: &logLevelValue];
	LOG_DEBUG("Result for %@: %d\n", PLIST_KEY_LOGGING_LEVEL, logLevelValue);
	LOG_DEBUG("  errorMessage=0x%p", errorMessage);
	if(errorMessage != nil) {
		[debugLoggingEnabledButton setState: NSOffState];		
		//[debugLoggingEnabledButton setEnabled: FALSE];
		//[errorMessages appendFormat: @"%@: %@\n", PLIST_KEY_LOGGING_LEVEL, errorMessage];
	}
	else if(logLevelValue == 0 || logLevelValue == 1)
		[debugLoggingEnabledButton setState: NSOffState];
	else if(logLevelValue == 2)
		[debugLoggingEnabledButton setState: NSOnState];		
	else
		[errorMessages appendFormat: @"%@ value (%d) invalid.\n", PLIST_KEY_LOGGING_LEVEL, logLevelValue];
	
	errorMessage = [DriverUtil getPreferenceValue: PLIST_KEY_ENABLE_CACHING outValue: &useUblioValue];
	LOG_DEBUG("Result for %@: %d\n", PLIST_KEY_ENABLE_CACHING, useUblioValue);
	LOG_DEBUG("  errorMessage=0x%p", errorMessage);
	if(errorMessage != nil) {
		[ublioEnabledButton setState: NSOffState];
		//[ublioEnabledButton setEnabled: FALSE];
		//[errorMessages appendFormat: @"%@: %@\n", PLIST_KEY_ENABLE_CACHING, errorMessage];
	}
	else if(useUblioValue == 0)
		[ublioEnabledButton setState: NSOffState];
	else if(useUblioValue == 1)
		[ublioEnabledButton setState: NSOnState];
	else
		[errorMessages appendFormat: @"%@ value (%d) invalid.\n", PLIST_KEY_ENABLE_CACHING, useUblioValue];
	
	errorMessage = [DriverUtil getPreferenceValue: PLIST_KEY_ENABLE_RAW_DEVICES outValue: &mountRdisksValue];
	LOG_DEBUG("Result for %@: %d\n", PLIST_KEY_ENABLE_RAW_DEVICES, mountRdisksValue);
	LOG_DEBUG("  errorMessage=0x%p", errorMessage);
	if(errorMessage != nil) {
		mountRdisksValue = 0;
		//[errorMessages appendFormat: @"%@: %@\n", PLIST_KEY_ENABLE_RAW_DEVICES, errorMessage];
	}
	if(mountRdisksValue == 0) {
		[ublioEnabledButton setState: NSOffState];
		[ublioEnabledButton setEnabled: FALSE];
	}
	else if(mountRdisksValue != 1)
		[errorMessages appendFormat: @"%@ value (%d) invalid.\n", PLIST_KEY_ENABLE_RAW_DEVICES, mountRdisksValue];
	
	errorMessage = [DriverUtil getPreferenceValue: PLIST_KEY_ENABLE_NORMALIZATION outValue: &useNfconvValue];
	LOG_DEBUG("Result for %@: %d\n", PLIST_KEY_ENABLE_NORMALIZATION, useNfconvValue);
	LOG_DEBUG("  errorMessage=0x%p", errorMessage);
	if(errorMessage != nil) {
		[nfconvEnabledButton setState: NSOffState];
		//[nfconvEnabledButton setEnabled: FALSE];
		//[errorMessages appendFormat: @"%@: %@\n", PLIST_KEY_ENABLE_NORMALIZATION, errorMessage];
	}
	else if(useNfconvValue == 0)
		[nfconvEnabledButton setState: NSOffState];
	else if(useNfconvValue == 1)
		[nfconvEnabledButton setState: NSOnState];
	else
		[errorMessages appendFormat: @"%@ value (%d) invalid.\n", PLIST_KEY_ENABLE_NORMALIZATION, useNfconvValue];
	
	errorMessage = [DriverUtil getPreferenceValue: PLIST_KEY_ENABLE_AUTOMATIC_RECOVERY outValue: &alwaysRecoverValue];
	LOG_DEBUG("Result for %@: %d\n", PLIST_KEY_ENABLE_AUTOMATIC_RECOVERY, alwaysRecoverValue);
	LOG_DEBUG("  errorMessage=0x%p", errorMessage);
	if(errorMessage != nil) {
		[alwaysRecoverButton setState: NSOffState];
		//[alwaysRecoverButton setEnabled: FALSE];
		//[errorMessages appendFormat: @"UseNfconv: %@\n", errorMessage];
	}
	else if(alwaysRecoverValue == 0)
		[alwaysRecoverButton setState: NSOffState];
	else if(alwaysRecoverValue == 1)
		[alwaysRecoverButton setState: NSOnState];
	/*else
		[errorMessages appendFormat: @"AlwaysRecover value (%d) invalid.\n", alwaysRecoverValue];*/
	
	errorMessage = [DriverUtil existsFile: DRIVER_UNINSTALL_SCRIPT_PATH
							   withResult: &existsUninstallScript
							   withBundle: [prefPane bundle]
								 withAuth: [[authView authorization] authorizationRef]];
	LOG_DEBUG("Result for existsUninstallScript: %d\n", existsUninstallScript);
	LOG_DEBUG("  errorMessage=0x%p", errorMessage);
	if(errorMessage != nil) {
		[uninstallDriverButton setEnabled: FALSE];
		[errorMessages appendFormat: @"[uninstallDriverButton] %@\n", errorMessage];
	}
	else if(existsUninstallScript == FALSE) {
		[uninstallDriverButton setEnabled: FALSE];
		[uninstallDriverButton setToolTip: @"Button disabled: Uninstall script not found!"];
	}
	
	if([errorMessages length] > 0)
 		[self errorMessage: errorMessages forWindow: [volumeList window]];
	
	LOG_DEBUG("Exiting refreshGlobalOptions...");
	
	[errorMessages release];
}

- (void)refreshVolumeOptions
{
	[self setVolumeOptionComponentsEnabled: FALSE];
	NSMutableArray *mountPointArray = [[NSMutableArray alloc] initWithCapacity: 10];
	NSMutableArray *statusArray = [[NSMutableArray alloc] initWithCapacity: 10];
	NSString *errorString =
	[DriverUtil listMountpoints: [[authView authorization] authorizationRef] toMountPointArray: mountPointArray toStatusArray: statusArray withBundle:[prefPane bundle]];
	if(errorString == nil) {
		[volumeList removeAllItems];
		if([mountPointArray count] != [statusArray count]) {
			NSString *msg = [[NSString alloc] initWithFormat: @"Internal error: [mountPointArray count](%d) != [statusArray count](%d)", [mountPointArray count], [statusArray count]];
			[self errorMessage: msg forWindow: [volumeList window]];
			[msg release];
		}
		else {
			/*
			int i;
			for(i = 0; i < [mountPointArray count]; ++i) {
				id curMountPoint = [mountPointArray objectAtIndex: i];
				id curStatus = [statusArray objectAtIndex: i];
				
				if([curStatus isEqual: @"  readonly"]) {
					NSString *roTitle = [[NSString alloc] initWithFormat: @"%@ (read-only)", curMountPoint];
					[volumeList addItemWithTitle: roTitle];
					[roTitle release];
				}
				else
					[volumeList addItemWithTitle: curMountPoint];
			}
			*/
			
			
			if([mountPointArray count] > 0) {
				[volumeList addItemsWithTitles: mountPointArray];
				[volumeList selectItemAtIndex: 0];
				[self selectVolume:self];
			}
		}
	}
	else {
		[self errorMessage: errorString forWindow: [volumeList window]];
	}
}

- (void)configureVolumeControlFile: (NSString*)controlFile forButton: (NSButton*)button
{
	long state = [button state];
	BOOL existent;
	if(state == NSOnState || state == NSOffState) {
		if(state == NSOnState)
			existent = TRUE;
		else if(state == NSOffState)
			existent = FALSE;
		
		NSString *selectedVolumePath = [volumeList titleOfSelectedItem];
		NSString *ntfs3gSettingDirPath = [[NSString alloc] initWithFormat: @"%@/%@", selectedVolumePath, DRIVER_CFDIRECTORY];
		
		NSString *errorMessage = [DriverUtil setControlFile: controlFile
												inDirectory: ntfs3gSettingDirPath
													toState: existent
												 withBundle: [prefPane bundle]
												   withAuth: [[authView authorization] authorizationRef]];
		[ntfs3gSettingDirPath release];
		
		if(errorMessage != nil) {
			[self errorMessage: errorMessage
					 forWindow: [button window]];
			
		}
		
		[self selectVolume: self];
	}
	else {
		NSString *msg = [[NSString alloc] initWithFormat: @"Invalid button state: %d", state];
		[self errorMessage: msg
				 forWindow: [button window]];
		[msg release];
	}
}

- (void)awakeFromNib
{
	[volumeList removeAllItems];
	//[debugField setStringValue: @"awoken from nibb!"];	
	
	[authView setDelegate: self];
	
	SFAuthorizationViewState vstate = [authView authorizationState];
	if(vstate == SFAuthorizationViewUnlockedState)
		[self setAllComponentsEnabled:TRUE];
	else {
		[self setAllComponentsEnabled:FALSE];
		//if(vstate == SFAuthorizationStartupState)
			//[debugField setStringValue:@"AuthorizationViewState: startup"];
		//else if(vstate == SFAuthorizationViewLockedState)
			//[debugField setStringValue:@"AuthorizationViewState: locked"];
		//else if(vstate == SFAuthorizationViewInProgressState)
			//[debugField setStringValue:@"AuthorizationViewState: in progress"];
	}
}

- (void)authorizationViewCreatedAuthorization:(SFAuthorizationView *)view
{
	//[debugField setStringValue:@"authorizationViewCreatedAuthorization"];
}

- (void)authorizationViewDidAuthorize:(SFAuthorizationView *)view
{
	//[debugField setStringValue:@"authorizationViewDidAuthorize"];
	[self setGlobalComponentsEnabled:TRUE];
	[self setVolumeControlComponentsEnabled:TRUE];
	
	[self refreshVolumeOptions];
	
	//[self errorMessage: @"hej" forWindow: [volumeList window]];
	[self refreshGlobalOptions];
	//[debugField setStringValue:@"authorizationViewDidAuthorize exiting"];
}

- (void)authorizationViewDidDeauthorize:(SFAuthorizationView *)view
{
	//[debugField setStringValue:@"authorizationViewDidDeauthorize"];
	[self setAllComponentsEnabled:FALSE];
}

- (void)authorizationViewReleasedAuthorization:(SFAuthorizationView *)view
{
	//[debugField setStringValue:@"authorizationViewReleasedAuthorization"];
}

- (BOOL)authorizationViewShouldDeauthorize:(SFAuthorizationView *)view
{
	//[debugField setStringValue:@"authorizationViewShouldDeauthorize"];
	return TRUE;
}

- (IBAction)clickToggleDriver:(id)sender
{
	// The view is the model :) (I'm lazy)
	NSString *currentDriverState = [driverEnabledStatusField stringValue];
	NSString *errorMessage;
	if([currentDriverState isEqualToString: @"Enabled"]) {
		// The user expects the driver to be set to a disabled state.
		//[debugField setStringValue:@"clickToggleDriver: "];
		errorMessage = [DriverUtil setPreferenceKey: PLIST_KEY_DISABLE_DRIVER
											toValue: @"1"
										 withBundle: [prefPane bundle]
										   withAuth: [[authView authorization] authorizationRef]];
	}
	else if([currentDriverState isEqualToString: @"Disabled"]) {
		// The user expects the driver to be set to an enabled state.
		errorMessage = [DriverUtil setPreferenceKey: PLIST_KEY_DISABLE_DRIVER
											toValue: @"0"
										 withBundle: [prefPane bundle]
										   withAuth: [[authView authorization] authorizationRef]];
	}
	else
		[self errorMessage: @"Undefined string value in driverEnabledStatusField"
				 forWindow: [driverEnabledStatusField window]];
	
	if(errorMessage != nil)
		[self errorMessage: errorMessage
				 forWindow: [driverEnabledStatusField window]];
	
	[self refreshGlobalOptions];
}

- (void) driverUninstalledAlertDidEnd:(NSAlert *)something returnCode:(int)somethingElse contextInfo:(void *)whatever
{
	[self restartOurselves];
	[self errorMessage: @"Could not restart System Preferences! Please close it manually. For information on what went wrong, check /var/log/system.log."
			 forWindow: [uninstallDriverButton window]];
}

- (IBAction)clickUninstallDriver:(id)sender
{
	//[debugField setStringValue:@"clickUninstallDriver"];
	
	NSMutableArray *output = [[NSMutableArray alloc] initWithCapacity: 0];
	NSString *errorMessage = [DriverUtil invokeUninstallScriptWithBundle: [prefPane bundle]
																withAuth: [[authView authorization]
																		   authorizationRef]
														   toOutputArray: output];
	
	if(errorMessage != nil) {
		NSMutableString *msg = [[NSMutableString alloc] initWithCapacity: 0];
		[msg appendFormat: @"%@\n\n", errorMessage];
		
		int i;
		for(i = 0; i < [output count]; ++i)
			[msg appendFormat: @"%@\n", [output objectAtIndex: i]];
		
		[self errorMessage: msg
				 forWindow: [uninstallDriverButton window]];
		
		[msg release];
	}
	else {
		[self messageWithTitle: @"Uninstall complete"
				  withInfoText: @"The System Preferences application will now be restarted."
				withAlertStyle: NSInformationalAlertStyle
					 forWindow: [uninstallDriverButton window]
				 modalDelegate: self
				didEndSelector: @selector(driverUninstalledAlertDidEnd:returnCode:contextInfo:)];
	}
	
	[output release];
}

- (IBAction)clickEnableUblio:(id)sender
{
	//[debugField setStringValue:@"clickEnableUblio"];
	
	NSString *errorMessage;
	if([ublioEnabledButton state] == NSOnState) {
		errorMessage = [DriverUtil setPreferenceKey: PLIST_KEY_ENABLE_CACHING
											toValue: @"1"
										 withBundle: [prefPane bundle]
										   withAuth: [[authView authorization] authorizationRef]];
	}
	else if([ublioEnabledButton state] == NSOffState) {
		errorMessage = [DriverUtil setPreferenceKey: PLIST_KEY_ENABLE_CACHING
											toValue: @"0"
										 withBundle: [prefPane bundle]
										   withAuth: [[authView authorization] authorizationRef]];
	}
	else
		[self errorMessage: @"Undefined button state for ublioEnabledButton."
				 forWindow: [ublioEnabledButton window]];
	
	if(errorMessage != nil)
		[self errorMessage: errorMessage
				 forWindow: [ublioEnabledButton window]];
		
	[self refreshGlobalOptions];
}

- (IBAction)clickEnableNfconv:(id)sender
{
	//[debugField setStringValue:@"clickEnableNfconv"];
	
	NSString *errorMessage;
	if([nfconvEnabledButton state] == NSOnState) {
		errorMessage = [DriverUtil setPreferenceKey: PLIST_KEY_ENABLE_NORMALIZATION
											toValue: @"1"
										 withBundle: [prefPane bundle]
										   withAuth: [[authView authorization] authorizationRef]];
	}
	else if([nfconvEnabledButton state] == NSOffState) {
		errorMessage = [DriverUtil setPreferenceKey: PLIST_KEY_ENABLE_NORMALIZATION
											toValue: @"0"
										 withBundle: [prefPane bundle]
										   withAuth: [[authView authorization] authorizationRef]];
	}
	else
		[self errorMessage: @"Undefined button state for nfconvEnabledButton."
				 forWindow: [nfconvEnabledButton window]];
	
	if(errorMessage != nil)
		[self errorMessage: errorMessage
				 forWindow: [nfconvEnabledButton window]];
	
	[self refreshGlobalOptions];	
}

- (IBAction)clickEnableDebugLogging:(id)sender
{
	//[debugField setStringValue:@"clickEnableDebugLogging"];
	
	NSString *valueString = nil;
	if([debugLoggingEnabledButton state] == NSOnState) {
		valueString = @"2";
	}
	else if([debugLoggingEnabledButton state] == NSOffState) {
		valueString = @"1";
	}
	else
		[self errorMessage: @"Undefined button state for debugLoggingEnabledButton."
				 forWindow: [debugLoggingEnabledButton window]];
	
	if(valueString != nil) {
		NSString *errorMessage = [DriverUtil setPreferenceKey: PLIST_KEY_LOGGING_LEVEL
													  toValue: valueString
												   withBundle: [prefPane bundle]
													 withAuth: [[authView authorization] authorizationRef]];
		
		if(errorMessage != nil) {
			[self errorMessage: errorMessage
					 forWindow: [debugLoggingEnabledButton window]];
		}
		
		[self refreshGlobalOptions];
	}
}

- (IBAction)clickAlwaysRecover:(id)sender
{
	//[debugField setStringValue:@"clickAlwaysRecover"];
	
	NSString *errorMessage;
	if([alwaysRecoverButton state] == NSOnState) {
		errorMessage = [DriverUtil setPreferenceKey: PLIST_KEY_ENABLE_AUTOMATIC_RECOVERY
											toValue: @"1"
										 withBundle: [prefPane bundle]
										   withAuth: [[authView authorization] authorizationRef]];
	}
	else if([alwaysRecoverButton state] == NSOffState) {
		errorMessage = [DriverUtil setPreferenceKey: PLIST_KEY_ENABLE_AUTOMATIC_RECOVERY
											toValue: @"0"
										 withBundle: [prefPane bundle]
										   withAuth: [[authView authorization] authorizationRef]];
	}
	else
		[self errorMessage: @"Undefined button state for alwaysRecoverButton."
				 forWindow: [alwaysRecoverButton window]];
	
	if(errorMessage != nil)
		[self errorMessage: errorMessage
				 forWindow: [alwaysRecoverButton window]];
	
	[self refreshGlobalOptions];	
}

- (IBAction)clickVolumeRefreshList:(id)sender
{
	//[debugField setStringValue:@"clickVolumeRefreshList"];
	[self refreshVolumeOptions];
}

- (IBAction)clickVolumeDisableUblio:(id)sender
{
	//[debugField setStringValue:@"clickVolumeDisableUblio"];
	
	[self configureVolumeControlFile: DRIVER_CF_NO_UBLIO forButton: volumeDisableUblioButton];
}

- (IBAction)clickVolumeDisableDecomposition:(id)sender
{
	//[debugField setStringValue:@"clickVolumeDisableDecomposition"];
	
	[self configureVolumeControlFile: DRIVER_CF_NO_NFCONV forButton: volumeDisableDecompositionButton];
}

- (IBAction)clickVolumeDisableDriver:(id)sender
{
	//[debugField setStringValue:@"clickVolumeDisableDriver"];
	
	[self configureVolumeControlFile: DRIVER_CF_READ_ONLY forButton: volumeDisableDriverButton];
}

- (IBAction)clickTuxeraInfoButton:(id)sender
{
	NSWorkspace *workspace = [NSWorkspace sharedWorkspace];
	NSURL *buyURL = [[NSURL alloc] initWithString: TUXERA_NTFS_INFO_URL];

	[workspace openURL: buyURL];
	[buyURL release];
}

- (IBAction)selectVolume:(id)sender
{
	NSString *errorMessage;

	//[debugField setStringValue:@"selectVolume"];
	
	NSString *selectedVolumePath = [volumeList titleOfSelectedItem];
	
	NSMutableArray *optionArray = [[NSMutableArray alloc] initWithCapacity: 0];
	errorMessage = [DriverUtil getMountFlags: selectedVolumePath toArray: optionArray withBundle: [prefPane bundle] withAuth: [[authView authorization] authorizationRef]];
	if(errorMessage != nil) {
		[self errorMessage: errorMessage
				 forWindow: [volumeDisableDriverButton window]];
	}
	else {
		BOOL isReadonlyMount = FALSE;
		{
			//NSMutableString *msg = [[NSMutableString alloc] initWithFormat: @"Mount options for \"%@\":\n", selectedVolumePath];
			int i;
			for(i = 0; i < [optionArray count]; ++i) {
				id currentObject = [optionArray objectAtIndex: i];
				//[msg appendFormat: @"  \"%@\"\n", currentObject];
				if([currentObject isEqual: @"MNT_RDONLY"] == TRUE) {
					//[msg appendFormat: @"    (isReadonlyMount flag was set)\n"];
					isReadonlyMount = TRUE;
					//break;
				}
			}
			//[self errorMessage: msg
			//		 forWindow: [volumeDisableDriverButton window]];
			//[msg release];
		}
		
		NSString *readonlyFilename = [[NSString alloc] initWithFormat: @"%@/%@/%@", selectedVolumePath, DRIVER_CFDIRECTORY, DRIVER_CF_READ_ONLY];
		NSString *noublioFilename =  [[NSString alloc] initWithFormat: @"%@/%@/%@", selectedVolumePath, DRIVER_CFDIRECTORY, DRIVER_CF_NO_UBLIO];
		NSString *nonfconvFilename = [[NSString alloc] initWithFormat: @"%@/%@/%@", selectedVolumePath, DRIVER_CFDIRECTORY, DRIVER_CF_NO_NFCONV];
		
		//[debugField setEnabled: TRUE];
		//[debugField setStringValue: readonlyFilename];
		
		[self setVolumeOptionComponentsEnabled: FALSE];
		
		BOOL res;
		if((errorMessage = [DriverUtil existsFile: readonlyFilename withResult: &res withBundle: [prefPane bundle] withAuth: [[authView authorization] authorizationRef]]) == nil) {
			if(res == TRUE)
				[volumeDisableDriverButton setState: NSOnState];
			else
				[volumeDisableDriverButton setState: NSOffState];
			
			if(!isReadonlyMount)
				[volumeDisableDriverButton setEnabled: TRUE];
		}
		else {
			[self errorMessage: errorMessage
					 forWindow: [volumeDisableDriverButton window]];
		}
		
		if((errorMessage = [DriverUtil existsFile: nonfconvFilename withResult: &res withBundle: [prefPane bundle] withAuth: [[authView authorization] authorizationRef]]) == nil) {
			if(res == TRUE)
				[volumeDisableDecompositionButton setState: NSOnState];
			else
				[volumeDisableDecompositionButton setState: NSOffState];
			
			if(!isReadonlyMount)
				[volumeDisableDecompositionButton setEnabled: TRUE];
		}
		else {
			[self errorMessage: errorMessage
					 forWindow: [volumeDisableDecompositionButton window]];
		}
		
		if((errorMessage = [DriverUtil existsFile: noublioFilename withResult: &res withBundle: [prefPane bundle] withAuth: [[authView authorization] authorizationRef]]) == nil) {
			if(res == TRUE)
				[volumeDisableUblioButton setState: NSOnState];
			else
				[volumeDisableUblioButton setState: NSOffState];
			
			if(!isReadonlyMount)
				[volumeDisableUblioButton setEnabled: TRUE];
		}
		else {
			[self errorMessage: errorMessage
					 forWindow: [volumeDisableUblioButton window]];
		}
		
		[nonfconvFilename release];
		[noublioFilename release];
		[readonlyFilename release];	
	}
	
	[optionArray release];
}

- (void)setGlobalOptionComponentsEnabled:(BOOL)enabled
{
	[ublioEnabledButton setEnabled:enabled];
	[nfconvEnabledButton setEnabled:enabled];
	[debugLoggingEnabledButton setEnabled:enabled];
	[alwaysRecoverButton setEnabled:enabled];
}

- (void)setGlobalComponentsEnabled:(BOOL)enabled
{
	[driverEnabledStatusField setEnabled: enabled];
	[toggleDriverButton setEnabled: enabled];
	[uninstallDriverButton setEnabled: enabled];
	[self setGlobalOptionComponentsEnabled: enabled];
}

- (void)setVolumeOptionComponentsEnabled:(BOOL)enabled
{
	[volumeDisableDriverButton setEnabled: enabled];
	[volumeDisableDecompositionButton setEnabled: enabled];
	[volumeDisableUblioButton setEnabled: enabled];
}

- (void)setVolumeControlComponentsEnabled:(BOOL)enabled
{
	[volumeList setEnabled: enabled];
	[volumeRefreshVolumeListButton setEnabled: enabled];
}

- (void)setVolumeComponentsEnabled:(BOOL)enabled
{
	[self setVolumeControlComponentsEnabled: enabled];
	[self setVolumeOptionComponentsEnabled: enabled];
}

- (void)setAllComponentsEnabled:(BOOL)enabled
{
	[self setGlobalComponentsEnabled: enabled];
	[self setVolumeComponentsEnabled: enabled];
	//[debugField setEnabled: enabled];
}

@end
