/*-
 * Copyright (C) 2008-2010 Erik Larsson
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

/*
 *  ProcessRunner.c
 *  NTFS-3G_PrefPane
 *
 *  Created by Erik Larsson on 2009-01-07.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
//#include <syslog.h>

#define error(msg, args...) \
	do { \
		fprintf(stderr, msg, ##args); \
		fprintf(stderr, "\n"); \
		exit(1); \
	} while(0)

int main(int argc, char **argv) {
	int retval = 1;

	//openlog("ProcessRunner", LOG_CONS | LOG_PID, LOG_USER);
	//openlog("ProcessRunner", LOG_CONS | LOG_PID | LOG_PERROR, LOG_USER);
	//syslog(LOG_NOTICE, "Started. uid=%d euid=%d", getuid(), geteuid());

	if(setuid(0) != 0)
		error("Could not setuid(0)!");
	else {
		//syslog(LOG_NOTICE, "After setuid: uid=%d euid=%d", getuid(), geteuid());
		if(argc < 2) {
			error("Invalid number of arguments. (got: %d, minimum required: 1)",
					(argc-1));
		}
		else {
			char *path = argv[1];

			int fildes[2];
			pipe(fildes);

			//syslog(LOG_NOTICE, "forking...");
			pid_t childPid = fork();
			if(childPid == -1) {
				//syslog(LOG_NOTICE, "ERROR: fork failed!");
				error("Could not fork()!");
			}
			else if(childPid == 0) {
				//syslog(LOG_NOTICE, "fork succeeded - now in child uid=%d euid=%d", getuid(), geteuid());
				//sleep(3);
				close(fildes[0]);
				fclose(stdout);
				dup2(fildes[1], STDOUT_FILENO);
				dup2(fildes[1], STDERR_FILENO);
				char **childArgs = argv+1;
				//syslog(LOG_NOTICE, "invoking execv...");
				execv(path, childArgs);
				//syslog(LOG_NOTICE, "ERROR: execv failed with errno=%d", errno);
				error("execv failed with errno=%d", errno);
				closelog();
				exit(errno);
			}
			else {
				int status;

				//syslog(LOG_NOTICE, "fork succeeded - now in parent");

				close(fildes[1]);
				FILE *pipe = fdopen(fildes[0], "r");
				if(pipe == NULL) {
					error("could not fdopen fildes[0]! errno=%d", errno);
					exit(retval);
				}
				char *curLine;
				size_t curLineLen;
				char *memBuf = calloc(1, 1);
				long memBufLen = 1;

				//syslog(LOG_NOTICE, "reading output from child process...");
				while((curLine = fgetln(pipe, &curLineLen)) != NULL) {
					//syslog(LOG_NOTICE, "  got line from child. length=%d memBufLen=%d", curLineLen, memBufLen);
					int curPos = memBufLen-1;

					//syslog(LOG_NOTICE, "  memBufLen (%d) incremented by curLineLen (%d)...", memBufLen, curLineLen);
					memBufLen += curLineLen;
					//syslog(LOG_NOTICE, "    result=%d", memBufLen);
					memBuf = realloc(memBuf, memBufLen);
					if(memBuf != NULL) {
						memcpy(memBuf+curPos, curLine, curLineLen);

						memBuf[memBufLen-1] = '\0';
					}
					else {
						//syslog(LOG_NOTICE, "ERROR(memBuf realloc failed for %d bytes! errno=%d)", memBufLen, errno);
						error("memBuf realloc failed for %d bytes! errno=%d", memBufLen, errno);
						exit(retval); // "panic"
					}
				}
				//syslog(LOG_NOTICE, "finished reading output from child process...");
				fclose(pipe);

				close(fildes[0]);

				//syslog(LOG_NOTICE, "waiting for child processes...");
				pid_t waitres = waitpid(childPid, &status, 0);
				//syslog(LOG_NOTICE, "waiting complete! waitres=%d", waitres);
				if(waitres > 0) {
					if(WIFEXITED(status)) {
						int exitValue = WEXITSTATUS(status);
						if(exitValue == 0) {
							retval = 0;
							if(memBufLen > 1)
								fprintf(stdout, "%s", memBuf);
						}
						else
							error("Non-zero exit value: %d\n%s", exitValue, memBuf);
					}
					else
						error("Process did not exit cleanly (crash?) status=0x%X\n%s", status, memBuf);
				}
				else
					error("waitpid failed! waitres=%d\n%s", waitres, memBuf);
			}
		}
	}

	//closelog();

	return retval;
}
