/*-
 * Copyright (C) 2008-2010 Erik Larsson
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

/*
 *  SetPlistValue.c
 *  NTFS-3G_PrefPane
 *
 *  Created by Erik Larsson on 2008-12-22.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <CoreFoundation/CoreFoundation.h>

#define error(msg, args...) \
	do { \
		fprintf(stderr, msg, ##args); \
		fprintf(stderr, "\n"); \
		exit(1); \
	} while(0)

CFPropertyListRef ReadPropertyListFromFile(CFURLRef fileURL)
{
	CFPropertyListRef propertyList;
	CFStringRef errorString;
	CFDataRef resourceData;
	Boolean status;
	SInt32 errorCode;

	// Read the XML file.
	status = CFURLCreateDataAndPropertiesFromResource(kCFAllocatorDefault,
													  fileURL,
													  &resourceData,            // place to put file data
													  NULL,
													  NULL,
													  &errorCode);

	// Reconstitute the dictionary using the XML data. (This method also works for binary
	// property list, even though the method is named as though only XML data was supported.)
	propertyList = CFPropertyListCreateFromXMLData(kCFAllocatorDefault,
												   resourceData,
												   kCFPropertyListImmutable,
												   &errorString);

	if(resourceData)
		CFRelease(resourceData);
	else
		CFRelease(errorString);

	return propertyList;
}

Boolean WritePropertyListToFile(CFPropertyListRef propertyList,
							 CFURLRef fileURL)
{
	CFDataRef xmlData;
	Boolean status;
	SInt32 errorCode;

	// Convert the property list into XML data.
	xmlData = CFPropertyListCreateXMLData(kCFAllocatorDefault, propertyList);

	// Write the XML data to the file.
	status = CFURLWriteDataAndPropertiesToResource(fileURL,                  // URL to use
												   xmlData,                  // data to write
												   NULL,
												   &errorCode);

	CFRelease(xmlData);

	return status;
}

int main(int argc, char **argv)
{
	if(argc != 4) {
		error("Invalid number of arguments. (got: %d, required: 3)",
				(argc-1));
	}
	else {
		CFStringRef cfFilenameArgument = CFStringCreateWithCString(kCFAllocatorDefault,
																   argv[1],
																   kCFStringEncodingUTF8);
		CFStringRef cfKeyArgument = CFStringCreateWithCString(kCFAllocatorDefault,
															  argv[2],
															  kCFStringEncodingUTF8);
		CFStringRef cfValueArgument = CFStringCreateWithCString(kCFAllocatorDefault,
																argv[3],
																kCFStringEncodingUTF8);

		CFURLRef fileURL = CFURLCreateWithFileSystemPath(kCFAllocatorDefault,
														 cfFilenameArgument,
														 kCFURLPOSIXPathStyle,
														 false);

		if(!cfFilenameArgument || !cfKeyArgument || !cfValueArgument || !fileURL)
			error("Failed to create necessary objects. cfFileNameArgument=%p cfKeyArgument=%p "
				  "cfValueArgument=%p fileURL=%p", cfFilenameArgument, cfKeyArgument, cfValueArgument, fileURL);

		CFPropertyListRef propertyList = ReadPropertyListFromFile(fileURL);
		if(propertyList == NULL)
			error("NULL property list.");
		else if(CFGetTypeID(propertyList) != CFDictionaryGetTypeID()) // "instanceof"
			error("Invalid property list root type: %d", CFGetTypeID(propertyList));
		else { 
			CFDictionaryRef settingsDict = (CFDictionaryRef) propertyList;

			CFMutableDictionaryRef newDict = CFDictionaryCreateMutableCopy(kCFAllocatorDefault, 0, settingsDict);
			CFDictionarySetValue(newDict, cfKeyArgument, cfValueArgument);

			if(!WritePropertyListToFile(newDict, fileURL)) {
				error("Failed to write property list to file \"%s\"", argv[1]);
			}
		}

		if(propertyList != NULL)
			CFRelease(propertyList);
		if(fileURL != NULL)
			CFRelease(fileURL);
		if(cfValueArgument != NULL)
			CFRelease(cfValueArgument);
		if(cfKeyArgument != NULL)
			CFRelease(cfKeyArgument);
		if(cfFilenameArgument != NULL)
			CFRelease(cfFilenameArgument);
	}

	return 0;
}

