/*-
 * Copyright (C) 2009 Erik Larsson / Tuxera Ltd.
 * 
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA
 */

/**
 * newfs_ntfs-3g
 *
 * A wrapper binary for the ntfs-3g.util shell script when doing a format
 * operation.
 * The reason that this is needed is that when a newfs operation is attempted,
 * it is done within some kind of setuid context that degrades the privileges of
 * ntfs-3g.util (Mac OS X doesn't allow setuid shell scripts). This means that
 * ntfs-3g.util doesn't get root privileges which it needs to access internal
 * disk devices.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

//#define SH_PATH "/bin/sh"

#ifdef TUXERA
#define UTIL_SCRIPT_PATH "/System/Library/Filesystems/fusefs_txantfs.fs/fusefs_txantfs.util"
#define LOG_FILE "/Library/Logs/newfs_fusefs_txantfs.log"
#else
#define UTIL_SCRIPT_PATH "/System/Library/Filesystems/ntfs-3g.fs/ntfs-3g.util"
#define LOG_FILE "/Library/Logs/newfs_ntfs-3g.log"
#endif

#define logToStderr(...)			\
  do {						\
    FILE *logFile = stderr;			\
    fprintf(logFile, __VA_ARGS__);		\
    fprintf(logFile, "\n");			\
  } while(0)
#define logToFile(...)							\
  do {									\
    FILE *logFile = fopen(LOG_FILE, "a");				\
    if(logFile != NULL) {						\
      fprintf(logFile, __VA_ARGS__);					\
      fprintf(logFile, "\n");						\
      fclose(logFile);							\
    }									\
    else {								\
      fprintf(stdout, "Could not open log file! %d (%s)",		\
	      errno, strerror(errno));					\
      exit(91);								\
    }									\
  } while(0)

#ifdef DEBUG
#define log(...)				\
  do {						\
    logToFile(__VA_ARGS__);			\
    logToStderr(__VA_ARGS__);			\
  } while(0)
#else
#define log(...)
#endif

#define error(...)				\
  do {						\
    logToFile(__VA_ARGS__);			\
    logToStderr(__VA_ARGS__);			\
  } while(0)

int main(int argc, char **argv) {
  int i = 0, j = 0;

  log("newfs_ntfs-3g:");
  log("  uid: %d", getuid());
  log("  euid: %d", geteuid());
  log("  argc: %d", argc);
  log("  invoked with:");
  for(i = 0; i < argc; ++i)
    log("    argv[%d]=\"%s\"", i, argv[i]);

  if(geteuid() == 0 && (getuid() != 0 || getgid() != 0)) {
    log("  setting uid to 0...");
    if(setuid(0) != 0)
      log("    failure: %d (%s)", errno, strerror(errno));
    else
      log("    ok.");

    log("  setting gid to 0...");
    if(setgid(0) != 0)
      log("    failure: %d (%s)", errno, strerror(errno));
    else
      log("    ok.");
  }

  const int extra_args = 2;
  int new_argv_size = extra_args + (argc-1) + 1;
  char **new_argv = calloc(1, new_argv_size * sizeof(char**));
  j = 0;

  //new_argv[j++] = SH_PATH;
  new_argv[j++] = UTIL_SCRIPT_PATH;
  new_argv[j++] = "-f";

  if(j != extra_args) {
    error("ASSERTION FAILED: j(%d) != extra_args(%d)", j, extra_args);
    return -1;
  }

  for(i = 1; i < argc; ++i)
    new_argv[j++] = argv[i];

  if(j != new_argv_size-1) {
    error("ASSERTION FAILED: j(%d) != new_argv_size-1(%d)", j, new_argv_size-1);
    return -1;
  }

  new_argv[j++] = NULL;

  execv(UTIL_SCRIPT_PATH, new_argv);

  int final_errno = errno;

  error("executing \"" UTIL_SCRIPT_PATH "\" failed:");
  error("  %d (%s)", final_errno, strerror(final_errno));
  error("  new_argv_size=%d", new_argv_size);
  for(i = 0; i < new_argv_size; ++i) {
    if(new_argv[i] == NULL)
      error("  new_argv[%d] = NULL", i);
    else
      error("  new_argv[%d] = \"%s\"", i, new_argv[i]);
  }

  return final_errno;
}
