/*-
 * Copyright (C) 2009 Erik Larsson
 * 
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA
 */

/**
 * ntfs-3g_confirm_mount
 *
 * A small helper tool which displays a CoreFoundation dialog notification to
 * ask the user for confirmation before attempting an NTFS-3G mount, and then,
 * if the user approves it, performs the mount.
 * It is used to ask the user whether it wants to force mount a hibernated or
 * uncleanly detached NTFS volume.
 */

#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <CoreFoundation/CoreFoundation.h>

#define PROGRAM_NAME "ntfs-3g_confirm_mount"
#define LOG_ENABLE_TRACE 0

#if LOG_ENABLE_TRACE
#define LOG_TRACE(...) do { FILE *logFile = stderr; fprintf(logFile, "[" PROGRAM_NAME "] DEBUG: "); fprintf(logFile, __VA_ARGS__); fprintf(logFile, "\n"); fflush(logFile); } while(0)
#else
#define LOG_TRACE(...)
#endif

/**
 * Prints usage instructions for this utility.
 *
 * @param outputStream the stream to which the instructions should be written (usually stdout or stderr)
 * @param invokeCmd the command used to invoke this utility (the utility's argv[0]).
 */
static void printUsage(FILE *outputStream, const char *invokeCmd) {
  /*                80: <--------------------------------------------------------------------------------> */
  fprintf(outputStream, "usage: %s\n", invokeCmd);
  fprintf(outputStream, "         <dialog title> <dialog message> <force button text>\n");
  fprintf(outputStream, "         <abort button text> <path to ntfs-3g binary> <device> <mountpoint>\n");
  fprintf(outputStream, "         <options>\n");
}

/**
 * Utility function that shows a two button CoreFoundation dialog box with the specified title,
 * body, default button text and alternate button text. This method blocks until the user has
 * responded to the dialog.
 *
 * The alertLevel parameter decides the seriousness of the notification and may have the values:
 *   kCFUserNotificationStopAlertLevel
 *   kCFUserNotificationNoteAlertLevel
 *   kCFUserNotificationCautionAlertLevel
 *   kCFUserNotificationPlainAlertLevel
 * 
 * If successful, a non-negative value is returned indicating the user's choice:
 *   0: User closed the dialog without choosing a button.
 *   1: User clicked the default button
 *   2: User clicked the alternate button
 * 
 * If unsuccessful, a negative value is returned and an error message gets printed to stderr.
 * Look in the code for the specific meaning of that value.
 */
static int showTwoButtonDialog(int alertLevel, const char *title, const char *body,
			       const char *defaultButtonText, const char *alternateButtonText) {
  int retval = -1;
  
  CFMutableDictionaryRef dialogElements =
    CFDictionaryCreateMutable(kCFAllocatorDefault, // Allocator
			      0, // Capacity
			      &kCFTypeDictionaryKeyCallBacks, // Default CFType key callbacks
			      &kCFTypeDictionaryValueCallBacks); // Default CFType value callbacks
    
  if(dialogElements == NULL) {
    fprintf(stderr, "ERROR: Failed to create dialogElements!\n");
    retval = -2;
  }
  else {
    CFStringRef cfTitle = NULL;
    CFStringRef cfMessage = NULL;
    CFStringRef cfDefaultButtonText = NULL;
    CFStringRef cfAlternateButtonText = NULL;
    
    cfTitle = CFStringCreateWithCString(kCFAllocatorDefault, title, kCFStringEncodingUTF8);
    cfMessage = CFStringCreateWithCString(kCFAllocatorDefault, body, kCFStringEncodingUTF8);
    cfDefaultButtonText = CFStringCreateWithCString(kCFAllocatorDefault, defaultButtonText, kCFStringEncodingUTF8);
    if(alternateButtonText != NULL)
      cfAlternateButtonText = CFStringCreateWithCString(kCFAllocatorDefault, alternateButtonText, kCFStringEncodingUTF8);
    
    if(cfTitle == NULL || cfMessage == NULL || cfDefaultButtonText == NULL ||
       (alternateButtonText != NULL && cfAlternateButtonText == NULL)) {
      fprintf(stderr, "ERROR: Failed to create one or more CFStrings! [%p, %p, %p, %p]\n",
	      cfTitle, cfMessage, cfDefaultButtonText, cfAlternateButtonText);
      retval = -3;
    }
    else {
      CFDictionaryAddValue(dialogElements, kCFUserNotificationAlertHeaderKey, cfTitle);
      CFDictionaryAddValue(dialogElements, kCFUserNotificationAlertMessageKey, cfMessage);
      CFDictionaryAddValue(dialogElements, kCFUserNotificationDefaultButtonTitleKey, cfDefaultButtonText);
      if(alternateButtonText != NULL)
	CFDictionaryAddValue(dialogElements, kCFUserNotificationAlternateButtonTitleKey, cfAlternateButtonText);
      
      SInt32 error;
      CFUserNotificationRef userNotification =
	CFUserNotificationCreate(kCFAllocatorDefault, // Allocator
				 0, // Timeout
				 alertLevel, // Flags
				 &error, // Error value, if any
				 dialogElements); // A dictionary describing the elements of the dialog
	
      if(userNotification == NULL) {
	fprintf(stderr, "ERROR: Failed to create userNotification!\n");
	retval = -4;
      }
      else {
	CFOptionFlags responseFlags;
	SInt32 responseResult =
	  CFUserNotificationReceiveResponse(userNotification,
					    0, // Timeout
					    &responseFlags);
	if(responseResult != 0) {
	  fprintf(stderr, "ERROR: CFUserNotificationReceiveResponse failed!\n");
	  retval = -5;
	}
	else {
	  int responseValue = responseFlags & 0x3;
	  switch(responseValue) {
	  case kCFUserNotificationCancelResponse:
	    retval = 0;
	    break;
	  case kCFUserNotificationDefaultResponse:
	    retval = 1;
	    break;
	  case kCFUserNotificationAlternateResponse:
	    if(alternateButtonText != NULL)
	      retval = 2;
	    else {
	      fprintf(stderr, "ERROR: Unexpected dialog response 'Alternate'!\n");
	      retval = -6;
	    }
	    break;
	  case kCFUserNotificationOtherResponse:
	    fprintf(stderr, "ERROR: Unexpected dialog response 'Other'!\n");
	    retval = -7;
	    break;
	  default:
	    fprintf(stderr, "ERROR: Invalid response value. THIS CAN'T HAPPEN!\n");
	    /* Really, it can't. But still... should the laws of the universe
	     * bend, we could at least print a message to stderr. ;) */
	    retval = -8;
	    break;
	  }
	}
	CFRelease(userNotification);
      }
    }
      
    if(cfAlternateButtonText != NULL)
      CFRelease(cfAlternateButtonText);
    if(cfDefaultButtonText != NULL)
      CFRelease(cfDefaultButtonText);
    if(cfMessage != NULL)
      CFRelease(cfMessage);
    if(cfTitle != NULL)
      CFRelease(cfTitle);
    
    CFRelease(dialogElements);
  }
  
  return retval;
}

/**
 * Utility function that shows a single button CoreFoundation dialog box with the specified title,
 * body and default button text. This method blocks until the user has responded to the dialog.
 * 
 * If successful, a non-negative value is returned indicating the user's choice:
 *   0: User closed the dialog without choosing a button.
 *   1: User clicked the default button
 * 
 * If unsuccessful, a negative value is returned and an error message gets printed to stderr.
 * Look in the code for the specific meaning of that value.
 */
static int showSingleButtonDialog(int alertLevel, const char *title, const char *body, const char *defaultButtonText) {
  return showTwoButtonDialog(alertLevel, title, body, defaultButtonText, NULL);
}

/**
 * Utility function which invokes the ntfs-3g binary located at ntfs3gPath with the specified
 * mount device, mount point and mount options. Before invoking ntfs-3g, it ensures that the
 * mount point exists, and creates it if it does not.
 *
 * Error messages are printed to stderr in case the mount is unsuccessful for some reason.
 *
 * @param ntfs3gPath      absolute path to the ntfs-3g executable to invoke.
 * @param mountDevicePath absolute path to the device to mount.
 * @param mountPointPath  absolute path to the mount point where the volume will be mounted.
 *                        If it does not exist, it will be created.
 * @param mountOptions    the mount options to pass to ntfs-3g (must begin with '-o').
 * @param ntfs3gOutput    the output of the ntfs-3g process stdout and stderr saved to a char
 *                        buffer.
 * @return 0 if the invocation of ntfs-3g was successful, and some other value if unsuccessful.
 */
static int attemptNtfs3gMount(char *ntfs3gPath, char *mountDevicePath, char *mountPointPath,
			      char *mountOptions, char **ntfs3gOutput) {
  DIR *dir;
  
  *ntfs3gOutput = NULL;
  
  dir = opendir(mountPointPath);
  if(dir == NULL) {
    printf("Creating mount point directory...\n");
    if(mkdir(mountPointPath, 0777) != 0) {
      fprintf(stderr, "ERROR: Could not create mount point!\n");
      return -1;
    }
    else
      printf("  done.\n");
  }
  else {
    printf("Mount point exists. Skipping directory creation.\n");
    closedir(dir);
  }
  
  /* Create pipe for reading the output from the child process. */
  int fildes[2];
  pipe(fildes);
  
  printf("Attempting a mount using the command line: \"%s\" \"%s\" \"%s\" \"%s\"...\n",
	 ntfs3gPath, mountDevicePath, mountPointPath, mountOptions);
  
  pid_t childPid = fork();
  if(childPid == -1) {
    fprintf(stderr, "ERROR: Could not fork!\n");
    return -1;
  }
  else if(childPid == 0) {
    // Child process code    
    int res;
    
    /* Close the read end of the pipe. */
    if((res = close(fildes[0])) != 0)
      fprintf(stderr, "WARNING: Child process could not close fildes[0]. res=%d errno=%d\n", res, errno);
    
    /* Connect the write end of the pipe to stdout and stderr of the child process. */
    if((res = dup2(fildes[1], STDOUT_FILENO)) != STDOUT_FILENO)
      fprintf(stderr, "WARNING: dup2 failed to connect fildes[1] to stdout. res=%d errno=%d\n", res, errno);
    if((res = dup2(fildes[1], STDERR_FILENO)) != STDERR_FILENO)
      fprintf(stderr, "WARNING: dup2 failed to connect fildes[1] to stderr. res=%d errno=%d\n", res, errno);
    
    char *exec_argv[5] = { ntfs3gPath, mountDevicePath, mountPointPath, mountOptions, NULL };
    res = execv(ntfs3gPath, exec_argv);
    
    fprintf(stderr, "ERROR: execv failed! res=%d, errno=%d\n", res, errno);
    close(fildes[1]);
    return -111;
  }
  else {
    // Parent process code
    int retval = -1;
    int res;

    /* Close the write end of the pipe. */
    if((res = close(fildes[1])) != 0)
      fprintf(stderr, "WARNING: Parent process could not close fildes[1]. res=%d errno=%d\n", res, errno);

    /* Wait for child process, if applicable. */
    int status;
    pid_t waitres = waitpid(childPid, &status, 0);
    if(waitres > 0) {
      if(WIFEXITED(status)) {
	int exitValue = WEXITSTATUS(status);
	if(exitValue == 0)
	  retval = 0;
	else {
	  fprintf(stderr, "ERROR: Non-zero exit value from forked process: %d\n", exitValue);
	  retval = exitValue;
	}
      }
      else
	fprintf(stderr, "ERROR: Process did not exit cleanly (crash?) status=0x%X\n", status);
    }
    else
      fprintf(stderr, "ERROR: waitpid failed! waitres=%d\n", waitres);
    
    if(retval != 0) {
      /* Read output from child process. */
      FILE *pipe = fdopen(fildes[0], "r");
      if(pipe != NULL) {
	char *curLine;
	size_t curLineLen;
	char *memBuf = calloc(1, 1);
	long memBufLen = 1;
	
	LOG_TRACE("Reading output from child process...");
	while((curLine = fgetln(pipe, &curLineLen)) != NULL) {
	  int curPos = memBufLen-1;
	  
	  LOG_TRACE("  Read line: %s", curLine);
	  
	  memBufLen += curLineLen;
	  memBuf = realloc(memBuf, memBufLen);
	  if(memBuf != NULL) {
	    memcpy(memBuf+curPos, curLine, curLineLen);
	    memBuf[memBufLen-1] = '\0';
	  }
	  else {
	    fprintf(stdout, "FATAL ERROR: memBuf realloc failed for %ld bytes! errno=%d\n", memBufLen, errno);
	    exit(retval); // "panic"
	  }
	}
	LOG_TRACE("Finished reading output from child process.");
	*ntfs3gOutput = memBuf;
	
	fclose(pipe);
      }
      else
	fprintf(stdout, "ERROR: Could not fdopen fildes[0]! errno=%d\n", errno);
    }
    
    close(fildes[0]);
    
    return retval;
  }
}

int main(int argc, char **argv) {
  int retval = -1;
  
  /*
  printf("STDOUT_FILENO=%d\n", STDOUT_FILENO);
  printf("STDERR_FILENO=%d\n", STDERR_FILENO);
  return 0;
  */
  
  /*
   * Accept the input in a specified order:
   * 1. <dialog title>
   * 2. <dialog message>
   * 3. <'force' button text>
   * 4. <'abort' button text>
   * 5. <path to ntfs-3g binary>
   * 6  <device to mount>
   * 7  <mountpoint>
   * 8. <ntfs-3g options>
   */
  if(argc != 9)
    printUsage(stderr, argv[0]);
  else {
    char *dialogTitle = argv[1];
    char *dialogMessage = argv[2];
    char *forceButtonText = argv[3];
    char *abortButtonText = argv[4];
    char *ntfs3gPath = argv[5];
    char *mountDevicePath = argv[6];
    char *mountPointPath = argv[7];
    char *mountOptions = argv[8];
    
    int dialogRes = showTwoButtonDialog(kCFUserNotificationCautionAlertLevel,
					dialogTitle, dialogMessage, abortButtonText, forceButtonText);
    if(dialogRes < 0) {
      fprintf(stderr, "ERROR: An error occurred while showing CoreFoundation dialog: %d\n", dialogRes);
    }
    else if(dialogRes == 0 || dialogRes == 1) {
      retval = 0;
    }
    else if(dialogRes == 2) {
      char *mountOutput;
      int mountRes = attemptNtfs3gMount(ntfs3gPath, mountDevicePath, mountPointPath, mountOptions, &mountOutput);
      if(mountRes == 0) {
	/* showSingleButtonDialog(kCFUserNotificationNoteAlertLevel, "NTFS-3G: Mount successful",
	 *                        "The mount completed successfully.", "Close");
	 */
      }
      else {
	showSingleButtonDialog(kCFUserNotificationStopAlertLevel, "NTFS-3G: Mount unsuccessful",
			       (mountOutput != NULL ? mountOutput : "NTFS-3G was unable to complete the mount."),
			       "Close");
      }
      
      retval = mountRes;
    }
    else {
      fprintf(stderr, "INTERNAL ERROR: Unrecognized showDialog return value: %d\n", dialogRes);
    }
  }

  return retval;
}
