/*-
 * Copyright (C) 2009 Erik Larsson
 * 
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA
 */

/**
 * showdialog
 *
 * A small helper tool which displays a CoreFoundation dialog with a specified
 * type, message, title and close button text.
 */

#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <libgen.h>
#include <sys/stat.h>
#include <CoreFoundation/CoreFoundation.h>

#ifdef DEBUG
#define LOG_ENABLE_TRACE 1
#else
#define LOG_ENABLE_TRACE 0
#endif

#if LOG_ENABLE_TRACE
#define LOG_TRACE(...) do { FILE *logFile = stderr; fprintf(logFile, "DEBUG: "); fprintf(logFile, __VA_ARGS__); fprintf(logFile, "\n"); fflush(logFile); } while(0)
#else
#define LOG_TRACE(...)
#endif

#define LOG_WARNING(...) do { FILE *logFile = stderr; fprintf(logFile, "WARNING: "); fprintf(logFile, __VA_ARGS__); fprintf(logFile, "\n"); fflush(logFile); } while(0)

#define LOG_ERROR(...) do { FILE *logFile = stderr; fprintf(logFile, "ERROR: "); fprintf(logFile, __VA_ARGS__); fprintf(logFile, "\n"); fflush(logFile); } while(0)

/**
 * Prints usage instructions for this utility.
 *
 * @param outputStream the stream to which the instructions should be written
 * (usually stdout or stderr)
 * @param invokeCmd the command used to invoke this utility (the utility's
 * argv[0]).
 */
static void printUsage(FILE *outputStream, const char *invokeCmd) {
  /*                80: <--------------------------------------------------------------------------------> */
  LOG_TRACE("Entering: printUsage(%p, \"%s\")", outputStream, invokeCmd);
  fprintf(outputStream, "usage: %s\n", invokeCmd);
  fprintf(outputStream, "         [stop|note|caution|plain] <dialog title> <dialog message>\n");
  fprintf(outputStream, "         <close button text>\n");
  LOG_TRACE("Leaving: printUsage(%p, \"%s\")", outputStream, invokeCmd);
}

/**
 * Utility function that shows a two button CoreFoundation dialog box with the
 * specified title, body, default button text and alternate button text. This
 * method blocks until the user has responded to the dialog.
 *
 * The alertLevel parameter decides the seriousness of the notification and may
 * have the values:
 *   kCFUserNotificationStopAlertLevel
 *   kCFUserNotificationNoteAlertLevel
 *   kCFUserNotificationCautionAlertLevel
 *   kCFUserNotificationPlainAlertLevel
 * 
 * If successful, a non-negative value is returned indicating the user's choice:
 *   0: User closed the dialog without choosing a button.
 *   1: User clicked the default button
 *   2: User clicked the alternate button
 * 
 * If unsuccessful, a negative value is returned and an error message gets
 * printed to stderr.
 * Look in the code for the specific meaning of that value.
 */
static int showTwoButtonDialog(int alertLevel, const char *title, const char *body,
			       const char *defaultButtonText, const char *alternateButtonText) {
  int retval = -1;
  
  CFMutableDictionaryRef dialogElements =
    CFDictionaryCreateMutable(kCFAllocatorDefault, // Allocator
			      0, // Capacity
			      &kCFTypeDictionaryKeyCallBacks, // Default CFType key callbacks
			      &kCFTypeDictionaryValueCallBacks); // Default CFType value callbacks
    
  if(dialogElements == NULL) {
    LOG_ERROR("Failed to create dialogElements!");
    retval = -2;
  }
  else {
    CFStringRef cfTitle = NULL;
    CFStringRef cfMessage = NULL;
    CFStringRef cfDefaultButtonText = NULL;
    CFStringRef cfAlternateButtonText = NULL;
    
    cfTitle = CFStringCreateWithCString(kCFAllocatorDefault, title,
					kCFStringEncodingUTF8);
    cfMessage = CFStringCreateWithCString(kCFAllocatorDefault, body,
					  kCFStringEncodingUTF8);
    cfDefaultButtonText = CFStringCreateWithCString(kCFAllocatorDefault,
						    defaultButtonText,
						    kCFStringEncodingUTF8);
    if(alternateButtonText != NULL)
      cfAlternateButtonText = CFStringCreateWithCString(kCFAllocatorDefault,
							alternateButtonText,
							kCFStringEncodingUTF8);
    
    if(cfTitle == NULL || cfMessage == NULL || cfDefaultButtonText == NULL ||
       (alternateButtonText != NULL && cfAlternateButtonText == NULL)) {
      LOG_ERROR("Failed to create one or more CFStrings! [%p, %p, %p, %p]",
	      cfTitle, cfMessage, cfDefaultButtonText, cfAlternateButtonText);
      retval = -3;
    }
    else {
      CFDictionaryAddValue(dialogElements, kCFUserNotificationAlertHeaderKey,
			   cfTitle);
      CFDictionaryAddValue(dialogElements, kCFUserNotificationAlertMessageKey,
			   cfMessage);
      CFDictionaryAddValue(dialogElements,
			   kCFUserNotificationDefaultButtonTitleKey,
			   cfDefaultButtonText);
      
      if(alternateButtonText != NULL)
	CFDictionaryAddValue(dialogElements,
			     kCFUserNotificationAlternateButtonTitleKey,
			     cfAlternateButtonText);
      
      SInt32 error;
      CFUserNotificationRef userNotification =
	CFUserNotificationCreate(kCFAllocatorDefault, // Allocator
				 0, // Timeout
				 alertLevel, // Flags
				 &error, // Error value, if any
				 dialogElements); // The elements of the dialog
	
      if(userNotification == NULL) {
	LOG_ERROR("Failed to create userNotification!");
	retval = -4;
      }
      else {
	CFOptionFlags responseFlags;
	SInt32 responseResult =
	  CFUserNotificationReceiveResponse(userNotification,
					    0, // Timeout
					    &responseFlags);
	if(responseResult != 0) {
	  LOG_ERROR("CFUserNotificationReceiveResponse failed!");
	  retval = -5;
	}
	else {
	  int responseValue = responseFlags & 0x3;
	  switch(responseValue) {
	  case kCFUserNotificationCancelResponse:
	    retval = 0;
	    break;
	  case kCFUserNotificationDefaultResponse:
	    retval = 1;
	    break;
	  case kCFUserNotificationAlternateResponse:
	    if(alternateButtonText != NULL)
	      retval = 2;
	    else {
	      LOG_ERROR("Unexpected dialog response 'Alternate'!");
	      retval = -6;
	    }
	    break;
	  case kCFUserNotificationOtherResponse:
	    LOG_ERROR("Unexpected dialog response 'Other'!");
	    retval = -7;
	    break;
	  default:
	    LOG_ERROR("Invalid response value. THIS CAN'T HAPPEN!");
	    /* Really, it can't. But still... should the laws of the universe
	     * bend, we could at least print a message to stderr. ;) */
	    retval = -8;
	    break;
	  }
	}
	CFRelease(userNotification);
      }
    }
      
    if(cfAlternateButtonText != NULL)
      CFRelease(cfAlternateButtonText);
    if(cfDefaultButtonText != NULL)
      CFRelease(cfDefaultButtonText);
    if(cfMessage != NULL)
      CFRelease(cfMessage);
    if(cfTitle != NULL)
      CFRelease(cfTitle);
    
    CFRelease(dialogElements);
  }
  
  return retval;
}

/**
 * Utility function that shows a single button CoreFoundation dialog box with
 * the specified title, body and default button text. This method blocks until
 * the user has responded to the dialog.
 * 
 * If successful, a non-negative value is returned indicating the user's choice:
 *   0: User closed the dialog without choosing a button.
 *   1: User clicked the default button
 * 
 * If unsuccessful, a negative value is returned and an error message gets
 * printed to stderr.
 * Look in the code for the specific meaning of that value.
 */
static int showSingleButtonDialog(int alertLevel, const char *title,
				  const char *body,
				  const char *defaultButtonText) {
  return showTwoButtonDialog(alertLevel, title, body, defaultButtonText, NULL);
}

int main(int argc, char **argv) {
  int retval = 1;
  
  /*
   * Accept the input in a specified order:
   * 1. <dialog type>
   * 2. <dialog title>
   * 3. <dialog message>
   * 4. <close button text>
   */
  
  if(argc != 5)
    printUsage(stderr, basename(argv[0]));
  else {
    char *dialogType = argv[1];
    char *dialogTitle = argv[2];
    char *dialogMessage = argv[3];
    char *closeButtonText = argv[4];
    
    int alertLevel;

    if(strcmp(dialogType, "stop") == 0)
      alertLevel = kCFUserNotificationStopAlertLevel;
    else if(strcmp(dialogType, "note") == 0)
      alertLevel = kCFUserNotificationNoteAlertLevel;
    else if(strcmp(dialogType, "caution") == 0)
      alertLevel = kCFUserNotificationCautionAlertLevel;
    else if(strcmp(dialogType, "plain") == 0)
      alertLevel = kCFUserNotificationPlainAlertLevel;
    else {
      LOG_ERROR("Invalid dialog type: %s", dialogType);
      printUsage(stderr, basename(argv[0]));
      return -1;
    }
    
    int dialogRes = showSingleButtonDialog(alertLevel, dialogTitle,
					   dialogMessage, closeButtonText);
    if(dialogRes < 0) {
      LOG_ERROR("An error occurred while showing CoreFoundation dialog: %d\n",
		dialogRes);
    }
    else if(dialogRes == 0 || dialogRes == 1) {
      retval = 0;
    }
    else {
      LOG_ERROR("Unrecognized showDialog return value: %d\n", dialogRes);
    }
  }

  return retval;
}
